%{
/*	makemenu.l	-- Greg Lee, lee@uhccux.uhcc.hawaii.edu
			   February 1989

	Generate C language menu module for Glib from menu specification.

Compile:
	flex makemenu.l
	cc -o makemenu lex.yy.c
or
	lex makemenu.l
	cc -o makemenu lex.yy.c -ll

Usage:
	makemenu <file.menu >file.c

Description:

Makemenu simplifies constructing and modifying synthesizer-specific
modules for Tim Thompson's Glib.  It allows the form of the on-screen
menu to be specified as a picture in the menu source file, and generates
'setval' and 'getval' calls automatically.

The following characterization of how makemenu works is probably
not entirely clear.  Look at the example d10wfg.menu and the
source code to get a more exact idea.  The source d10wfg.menu
should give a C file equivalent to d10wfg.c in the d10 version
of Glib done by Michael Kesti.

The menu source file should contain five special preprocessor
directives: '#MENU' (once), '#END' (once), '#O' (one for each
label name), '#GETVAL' (once), and '#SETVAL' (once).  The source
file has four sections:
	(1) the portion before the '#MENU' directive is passed
	    through unchanged to the output,
	(2) a picture of the screen menu is given between the
	    '#MENU' and '#END' directives,
	(3) after the '#END' directive is C code with '#O' directives
	    which specify the data offsets of variables and with
	    '%%' or '@@' references to the screen locations for
	    variable values, and
	(4) more C code containing somewhere the '#SETVAL' and
	    '#GETVAL' directives which cause a series of calls
	    to the 'setval' and 'getval' functions to be generated,
	    making use of the offset information given by the '#O'
	    directives in the preceding section.

The picture of the screen menu has constant strings in the position
they will be output to the screen and '%' or '@' in each position
where the current value of some variable is to be displayed by
Glib.  The locations of the '%'s and '@'s are remembered in the
order in which they are encountered (scanning left-to-right and
top-to-bottom in the picture), then the corresponding row/column
coordinates are substituted for '%%'s and '@@'s encountered in
section 3 of the source file.  The '%'s and '@'s are stored
independently so as to make it more convenient to specify menus
that have two columns of values.

The '#O' specifications are given in the form:
#O <integer-offset> "<label-name>" ...
and the quoted label-name is passed through unchanged to the
output.

*/
#define MAXLINE 133
#define MAXVAL 100
#define MAXVSTRINGS 200
#define MAXBUF 2000

	int i, j, k, wcol=0, col=0, row=0;
	char word[MAXLINE];
	int wordp = 0;
	int valpos[MAXVAL][2];
	int valpos2[MAXVAL][2];
	int valcnt = 0;
	int valcnt2= 0;
	int offset;
	int nlist[MAXVSTRINGS][2];
	int np = 0;
	char buf[MAXBUF];
	int bufp=0;
%}

%s M IG VAL VSTRING

%%

<IG>^#[ \t]*MENU.*\n	BEGIN(M);
<IG>\n			ECHO;
<IG>.			ECHO;

<M>"@"	{	writeword();
		if (valcnt2+1 >= MAXVAL) {
		    fprintf(stderr,"makemenu: too many @s in menu\n");
		    exit(1);
		}
		valpos2[valcnt2][0] = row;
		valpos2[valcnt2++][1] = col;
		col += 1;
};

<M>"%"	{	writeword();
		if (valcnt+1 >= MAXVAL) {
		    fprintf(stderr,"makemenu: too many %%s in menu\n");
		    exit(1);
		}
		valpos[valcnt][0] = row;
		valpos[valcnt++][1] = col;
		col += 1;
};

<M>[ ]*	{	if (yyleng > 4) writeword();
		else if (wordp) for (i=0; i<yyleng; i++) append(' ');
		col += yyleng;
};

<M>\t  {	writeword(); col = ((col + 8)/8)*8;
};

<M>\n  {	writeword(); col = 0; row += 1;
};

<M>.  {		append(yytext[0]); col += 1;
};

<M>^#[ \t]*END.*\n  {
		BEGIN(VAL); i = 0; j = 0;
};

<VAL>^"#O "[0-9]+[ \t\n]*  {
		offset = atoi(yytext+3);
		BEGIN(VSTRING);
};

<VSTRING>\"[^"\n]+/\"  {
		if (np+1 >= MAXVSTRINGS) {
		    fprintf(stderr,"makemenu: too many VSTRINGs\n");
		    exit(1);
		}
		if (bufp+yyleng >= MAXBUF) {
		    fprintf(stderr,"makemenu: out of storage room\n");
		    exit(1);
		}
		nlist[np][1] = offset;
		nlist[np++][0] = bufp;
		strcpy(buf+bufp, yytext+1);
		bufp += yyleng;
		ECHO;
		BEGIN(VAL);
};

<VSTRING>[^"]*	fprintf(stderr,"\nmakemenu: #O spec not followed by string\n");

<VAL>^#[ \t]*SETVAL.*\n  {
		for (k=0; k<np; k++)
		    printf("setval(\"%s\",data[RESERVESIZE + %d]);\n",
			buf+nlist[k][0], nlist[k][1]);
};

<VAL>^#[ \t]*GETVAL.*\n  {
		for (k=0; k<np; k++)
		    printf("data[RESERVESIZE + %d] = getval(\"%s\");\n",
			nlist[k][1], buf+nlist[k][0]);
};

<VAL>\\%%+	printf("%s", yytext+1);
<VAL>\\@@+	printf("%s", yytext+1);

<VAL>"@@"  {	if (j < valcnt2) {
			printf("%d, %d", valpos2[j][0], valpos2[j][1]);
			j++;
		}
		else fprintf(stderr,"\nmakemenu: too many @@s\n");
};

<VAL>"%%"  {	if (i < valcnt) {
			printf("%d, %d", valpos[i][0], valpos[i][1]);
			i++;
		}
		else fprintf(stderr,"\nmakemenu: too many %%%%s\n");
};

<VAL>.		ECHO;
<VAL>\n		ECHO;

%%
main()
{
	word[0] = 0;
	BEGIN(IG);
	yylex();
	if (i+1 < valcnt) fprintf(stderr,"\nmakemenu: not enough %%%%s\n");
	if (j+1 < valcnt2) fprintf(stderr,"\nmakemenu: not enough @@s\n");
	exit(0);
}

append(c)
char c;
{
	if (wordp+1 >= MAXLINE) {
	    fprintf(stderr,"makemenu: string in menu too long\n");
	    exit(1);
	}
	if (!wordp) wcol = col;
	word[wordp++] = c;
	word[wordp] = 0;
}

writeword()
{
	if (wordp) {
		while (word[--wordp] == ' ') ;
		word[++wordp] = 0;
		printf("%d,%d,\"%s\",\n", row, wcol, word);
		wordp = 0;
		word[wordp] = 0;
	}
}
